#include <cmath>
#include <cstdlib>
#include "algos.h"

#ifdef MT_SHARK
#  define Array MT::Array
#endif


//===========================================================================
//
// Linear Statistics
//

void LinearStatistics::compute(){
  if(computed) return;
  CHECK(accum,"no data accumulated in linear statistics");
  MeanX=meanX/accum;
  MeanY=meanY/accum;
  VarX =!(varX  - !(meanX^meanX)/accum)/accum;
  CovXY=!(covXY - !(meanX^meanY)/accum)/accum;
  computed=true;
}

void LinearStatistics::computeZeroMean(){
  if(computed) return;
  CHECK(accum,"no data accumulated in linear statistics");
  MeanX.resize(meanX); MeanX.setZero();
  MeanY.resize(meanY); MeanY.setZero();
  VarX =varX  /(accum*accum);
  CovXY=covXY /(accum*accum);
  computed=true;
}

void LinearStatistics::learn(const arr& X,const arr& Y,double weight){
  CHECK((X.nd==1 && Y.nd==1) || (X.nd==2 && Y.nd==2),"data must either be batch or online data");
  if(X.nd==1 && Y.nd==1){
    if(!weight) return;
    computed=false;
    if(!accum){ //first data received
      accum=weight;
      meanX=weight*X;
      meanY=weight*Y;
      varX =weight*!(X^X);
      covXY=weight*!(X^Y);
      return;
    }
    CHECK(samedim(X,meanX) && samedim(Y,meanY),"PLS: data dimensions differ");
    uint i,j;
    if(lambda>0.){
      accum=(1.-lambda*weight)*accum + weight;
      meanX=(1.-lambda*weight)*meanX + weight*X;
      meanY=(1.-lambda*weight)*meanY + weight*Y;
      FOR2D(varX,i,j) varX(i,j) =(1.-lambda*weight)*varX(i,j)  + weight*X(i)*X(j);
      covXY=(1.-lambda*weight)*covXY + weight*!(X^Y);
    }else{
      accum+=weight;
      FOR1D(meanX,i  ) meanX(i)  +=weight*X(i);
      FOR1D(meanY,i  ) meanY(i)  +=weight*Y(i);
      FOR2D(varX, i,j)  varX(i,j)+=weight*X(i)*X(j);
      FOR2D(covXY,i,j) covXY(i,j)+=weight*X(i)*Y(j);
    }
  }
  if(X.nd==2 && Y.nd==2){
    CHECK(X.d0==Y.d0,"need same number of in/out samples");
    arr ones(X.d0); ones=1.;
    computed=false;
    
    if(!accum){
      accum=X.d0;
      meanX=ones*X;
      meanY=ones*Y;
      covXY=!~X*Y;
      varX =!~X*X;
    }else{
      accum+=X.d0;
      meanX+=ones*X;
      meanY+=ones*Y;
      covXY+=!~X*Y;
      varX +=!~X*X;
    }
  }
}

void LinearStatistics::forget(double lambda){
  computed=false;
  accum=(1.-lambda)*accum;
  meanX=(1.-lambda)*meanX;
  meanY=(1.-lambda)*meanY;
  varX =(1.-lambda)*varX;
  covXY=(1.-lambda)*covXY;
}
void MT::normalizeData(arr& X){
  CHECK(X.nd==2,"data has to be a 2D batch");
  uint n,N=X.d0,k,K=X.d1;
  arr mean,var,x,sd;
  arr ones(N); ones=1.;

  innerProduct(mean,ones,X);
  mean/=(double)N;
  for(n=0;n<N;n++) X[n]()-=mean;

  innerProduct(var,!~X,X);
  var/=(double)N;
  sd.resize(K);
  for(k=0;k<K;k++) sd(k)=sqrt(var(k,k));
  for(n=0;n<N;n++) for(k=0;k<K;k++) if(sd(k)) X(n,k)/=sd(k);
}

//---------------------------------------------------------------------------
// code taken from jab@gk.dtu.dk: 
// 
// xspline.cc
//
// this file contains a class XSpline that implements the general XSplines
// presented by Blanc and Schlick in the article 
//
// ``A Spline Model Designed for the End-User.''
//
// Comments and bug-reports to jab@gk.dtu.dk
//---------------------------------------------------------------------------
void XSpline::eval(double t,arr& x,arr* v){
  CHECK(t>=0 && t<=V.d0-1,"out of range");

  //standard interpolation type (weighting of vertices)
  if(!W.N) type(true,1.);

  // query t is between two vertices V_{k+1} and V_{k+2}
  // the four vertices V_{k,..,k+3} will be used to calculate the output
  double kf = floor(t) - 1.0;
  int k = int(kf);

  // the weights associated to the four vertices are
  double W1,W2;
  W1=W(k+1);
  W2=k+2<(int)W.N?W(k+2):0.;
  
  // Find the big T values
  // 
  // T0p is the point where fk decreases to 0
  // T1p is the point where fk+1 decreases to 0
  // T2m is the point where fk+2 increases from 0
  // T3m is the point where fk+3 increases from 0
  //
  // The parameter sk of a control point k affects
  // the functions that have max vals in k-1 and k+1
  // Therefore, sv[k+1] affects T0p (The point where
  // fk decreases to 0) and T2m. Similarly, sv[k+2]
  // affects T1p and T3m
  double T0p = kf+1 + (W1>0?W1:0) * DELTA;
  double T1p = kf+2 + (W2>0?W2:0) * DELTA;
  double T2m = kf+1 - (W1>0?W1:0) * DELTA;
  double T3m = kf+2 - (W2>0?W2:0) * DELTA;
  
  // Find the pk values
  // The pm values are derived directly from the Ts above
  double tmp_factor = 2/(DELTA*DELTA);
  double pm1 = intpow(kf   - T0p, 2) * tmp_factor; 
  double pm0 = intpow(kf+1 - T1p, 2) * tmp_factor; 
  double pp1 = intpow(kf+2 - T2m, 2) * tmp_factor; 
  double pp2 = intpow(kf+3 - T3m, 2) * tmp_factor; 
  
  // The sv indices are determined like for the Ts
  double qp0 = W1<0?-W1/2.0:0;
  double qp1 = W2<0?-W2/2.0:0;
  double qp2 = W1<0?-W1/2.0:0;
  double qp3 = W2<0?-W2/2.0:0;
  
  
  // The code below resembles the part of (17) in the article, 
  // where A0 ... A3 are calculated, but there are differences
  // 
  // 1.
  // 
  // We use g instead of f. This is ok, because 
  //    q = 0 => g(u) = f(u) 
  //
  // 2.
  // we don't check whether we are inside the legal range 
  // of fk+1 and fk+2. This is not necessary, since we are in the
  // range [ tk+1, tk+2 ] and both gk+1 and gk+2 must be defined 
  // in this range.
  // 
  // 3.
  //
  // if we are in the ranges where fk and fk+3 would have been
  // nil in the case of basic splines, we use h instead of g
  // except if q
  //
  double A0,A1,A2,A3;      
  if(t<=T0p) A0 = g((t-T0p)/(kf-T0p), qp0, pm1);
  else       A0 = qp0>0 ? h((t-T0p)/(kf-T0p), qp0) : 0;
  A1 = g((t-T1p)/(kf+1-T1p), qp1, pm0);
  A2 = g((t-T2m)/(kf+2-T2m), qp2, pp1);
  if(t>=T3m) A3 = g((t-T3m)/(kf+3-T3m), qp3, pp2);
  else       A3 = qp3>0 ? h((t-T3m)/(kf+3-T3m), qp3) : 0;
  
  CHECK(k>=0 || A0==0.,"non-zero weight for out-of-range!");
  CHECK(k+3<(int)V.d0 || A3==0.,"non-zero weight for out-of-range!");

  double SUM = A0+A1+A2+A3;
  A0/=SUM;
  A1/=SUM;
  A2/=SUM;
  A3/=SUM;

  x.resize(V.d1); x.setZero();
  if(A0) x+=A0*V[k];
  if(A1) x+=A1*V[k+1];
  if(A2) x+=A2*V[k+2];
  if(A3) x+=A3*V[k+3];

  if(v){
    double dA0,dA1,dA2,dA3;      
    if(t<=T0p) dA0 = dgdu((t-T0p)/(kf-T0p), qp0, pm1)/(kf-T0p);
    else       dA0 = qp0>0 ? dhdu((t-T0p)/(kf-T0p), qp0)/(kf-T0p) : 0;
    dA1 = dgdu((t-T1p)/(kf+1-T1p), qp1, pm0)/(kf+1-T1p);
    dA2 = dgdu((t-T2m)/(kf+2-T2m), qp2, pp1)/(kf+2-T2m);
    if(t>=T3m) dA3 = dgdu((t-T3m)/(kf+3-T3m), qp3, pp2)/(kf+3-T3m);
    else       dA3 = qp3>0 ? dhdu((t-T3m)/(kf+3-T3m), qp3)/(kf+3-T3m) : 0;

    CHECK(k>=0 || dA0==0.,"non-zero weight for out-of-range!");
    CHECK(k+3<(int)V.d0 || dA3==0.,"non-zero weight for out-of-range!");

    double dSUM = dA0+dA1+dA2+dA3;
    dA0 = dA0/SUM - A0/SUM*dSUM;
    dA1 = dA1/SUM - A1/SUM*dSUM;
    dA2 = dA2/SUM - A2/SUM*dSUM;
    dA3 = dA3/SUM - A3/SUM*dSUM;

    v->resize(V.d1); v->setZero();
    if(dA0) *v+=dA0*V[k];
    if(dA1) *v+=dA1*V[k+1];
    if(dA2) *v+=dA2*V[k+2];
    if(dA3) *v+=dA3*V[k+3];
  }
}

void PartialLeastSquares::SIMPLS(){
  uint I=S.meanX.N,O=S.meanY.N,K=I,i;
  if(maxProj) K=maxProj;
  arr A,M,C;
  
  S.compute();
  double scale=trace(S.VarX);
  if(scale<1e-10){
    //MT_MSG("Warning: PLS: too small variance (only 1 data point?) -> use constant regression");
    B.resize(O,I); W.resize(K,I); Q.resize(K,O); B=0.; W=0.; Q=0.;
    return;
  }
  M=S.VarX/scale;
  A=S.CovXY/scale;
  C.resize(I,I); C=0.; for(i=0;i<I;i++) C(i,i)=1.;
  
  arr AA,eigU,eigw,eigV;
  double c;
  arr w,q,p,P,v;
  W.resize(K,I); P.resize(K,I); Q.resize(K,O); W=0.; Q=0.;
  resErr.resize(K); resErr=0.;
  for(uint k=0;k<K;k++){
    //step 0: check non-zero residuals:
    //std::cout <<sumOfSqr(A) <<std::endl;
    if(sumOfSqr(A)<1e-10) break;
    //step 1
    AA=!~A*A;
    svd(AA,eigU,eigw,eigV,true);
    q=eigV[0];
    //resErr(k)=eigw(0);
    //step 2
    w=A*q;
    c=scalarProduct(w,!(M*w));
    if(fabs(c)<1e-10) break;
    if(!(c>0.)) HALT("PLS: variance matrix M is not positive definite!");
    w/=sqrt(c);
    W[k]=w;
    //step 3
    p=M*w;
    P[k]=p;
    //step 4
    q=!~A*w;
    Q[k]=q;
    //step 5
    v=C*p; v/=norm(v);
    //step 6
    C-=!(v^v);
    M-=!(p^p);
    //step 7
    A=!(C*A);
  }
  //std::cout <<sumOfSqr(A) <<std::endl;
  B=!~Q*W;
}

void MT::makeSpline(arr& X,arr& P,uint intersteps){
  CHECK(P.nd==2,"makeSpline: set of points is 2D array");
  XSpline S;
  S.referTo(P);
  S.type(false,1.); //is default
  
  X.resize(1+(P.d0-1)*intersteps,P.d1);
  double tau;
  for(uint t=0;t<X.d0;t++){
    tau=((double)t)/intersteps;
    S.eval(tau,X[t]());
  }
}
void MT::makeSpline(arr& X,arr& dX,arr& P,uint intersteps){
  XSpline S;
  S.referTo(P);
  S.type(false,1.); //is default
  
  X.resize(1+(P.d0-1)*intersteps,P.d1);
  dX.resize(X);
  double tau;
  for(uint t=0;t<X.d0;t++){
    tau=((double)t)/intersteps;
    S.eval(tau,X[t](),dX[t]());
  }
}
void MT::randomSpline(arr& X,uint dim,uint points,uint intersteps,double lo,double hi,uint cycles){
  arr P(points,dim);
  rndUni(P,lo,hi);
  if(cycles>1) P.replicate(cycles);
  P.reshape(points*cycles,dim);
  P[0]() = 0.;
  makeSpline(X,P,intersteps);
}
void MT::randomSpline(arr& X,arr& dX,uint dim,uint points,uint intersteps,double lo,double hi,uint cycles){
  arr P(points,dim);
  rndUni(P,lo,hi);
  if(cycles>1) P.replicate(cycles);
  P.reshape(points*cycles,dim);
  P[0]() = 0.;
  makeSpline(X,dX,P,intersteps);
}

void checkGradient(void (*f)(arr&,const arr&),
		   void (*df)(arr&,const arr&),
		   arr& x,double tolerance=1e-5){
  arr y,J,dx,dy,JJ;
  f(y,x);
  df(J,x);

  uint n=x.N,m=y.N;
  JJ.resize(m,n);
  double eps=1e-6;
  uint i,k;
  for(i=0;i<x.N;i++){
    dx=x;
    dx.elem(i) += eps;
    f(dy,dx);
    dy = (dy-y)/eps;
    for(k=0;k<y.N;k++) JJ(k,i)=dy.elem(k);
  }
  double md=maxDiff(J,JJ);
  CHECK(md<tolerance,"FAILURE -- \n measured grad=" <<JJ <<"\ncomputed grad=" <<J <<"\nmax diff=" <<md);
  cout <<"checkGradient -- SUCCESS (max error=" <<md <<")" <<endl;
}

void checkGradient(double (*f)(const arr&),
		   void (*df)(arr&,const arr&),
		   arr& x,double tolerance=1e-5){
  arr J,dx,JJ;
  double y,dy;
  y=f(x);
  df(J,x);

  uint n=x.N;
  JJ.resize(n);
  double eps=1e-6;
  uint i;
  for(i=0;i<x.N;i++){
    dx=x;
    dx.elem(i) += eps;
    dy=f(dx);
    dy = (dy-y)/eps;
    JJ(i)=dy;
  }
  double md=maxDiff(J,JJ);
  CHECK(md<tolerance,"FAILURE -- \n measured grad=" <<JJ <<"\ncomputed grad=" <<J <<"\nmax diff=" <<md);
  cout <<"checkGradient -- SUCCESS" <<endl;
}

namespace minimizeStatic{
  uint n,fc,dfc;
  arr ff;
  double (*f)(const arr &x);
  void (*df)(arr &dx,const arr &x);
  double _f(double _x[]){
    arr x(_x,n);
    fc++;
    double y=f(x);
    ff.append(y);
    return y;
  }
  void _df(double _x[], double _dx[]){
    arr x(_x,n),dx(_dx,n);
    dfc++;
    df(dx,x);
  }
}
int minimize(double (*f)(const arr &x),void (*df)(arr &dx,const arr &x),arr &x,double tol=1e-3,bool chkGrad=true,bool saveCurve=true){
  void frprmn(double p[], int n, double ftol, int *iter, double *fret,
	      double (*func)(double []), void (*dfunc)(double [], double []));
  minimizeStatic::f=f;
  minimizeStatic::df=df;
  minimizeStatic::n=x.N;
  minimizeStatic::fc=0;
  minimizeStatic::dfc=0;
  double fmin;
  int i;
  if(chkGrad) checkGradient(f,df,x,tol);
  frprmn(x.p,x.N,tol,&i,&fmin,minimizeStatic::_f,minimizeStatic::_df);
  cout
    <<"--- minimization using conjugate gradient:"
    <<"\n  #iterations=" <<i
    <<"\n  f-counts="  <<minimizeStatic::fc
    <<"\n  df-counts=" <<minimizeStatic::dfc
    <<"\n  f(x_min)="  <<fmin
    <<"\n  x_min="     <<x
    <<endl;
  if(saveCurve){
    MT::IOraw=true;
    MT::save(~minimizeStatic::ff,"z.conjGrad");
  }
  return i;
}
