#ifndef MT_algos_h
#define MT_algos_h

#include "array.h"

namespace MT{

  void normalizeData(arr& X);
  void makeSpline(arr& X,arr& P,uint intersteps);
  void makeSpline(arr& X,arr& V,arr& P,uint intersteps);
  void randomSpline(arr& X,uint dim,uint points,uint intersteps=100,double lo=-1.,double hi=1.,uint cycles=1);
  void randomSpline(arr& X,arr& dX,uint dim,uint points,uint intersteps=100,double lo=-1.,double hi=1.,uint cycles=1);


  inline void eigensymm(const arr& A,arr& V,arr& w){
    arr U;
    ::svd(A,U,w,V,true);
  }


  inline void convolution(arr &y,const arr &x,double (*h)(double),double scale=1.){
    CHECK(x.nd==1 || x.nd==2,"");
    uint T=x.d0,i,j;
    y.resize(x); y.setZero();
    for(i=0;i<T;i++) for(j=0;j<T;j++){
      if(x.nd==1)
	y(i) += h(MT_2PI*((double)i-j)/scale)*2./scale * x(j);
      else
	y[i]() += h(MT_2PI*((double)i-j)/scale)*2./scale * x[j];
    }
  }
  inline void bandpassFilter(arr &y,const arr &x,double loWavelength,double hiWavelength){
    arr y_hi,y_lo;
    convolution(y_lo,x,MT::cosc,hiWavelength);
    convolution(y_hi,x,MT::cosc,loWavelength);
    y=y_hi-y_lo;
  }
  inline void bandpassEnergy(arr &y,const arr &x,double loWavelength,double hiWavelength){
    arr y_his,y_hic,y_los,y_loc,ys,yc;
    convolution(y_los,x,MT::sinc,hiWavelength);
    convolution(y_loc,x,MT::cosc,hiWavelength);
    convolution(y_his,x,MT::sinc,loWavelength);
    convolution(y_hic,x,MT::cosc,loWavelength);
    ys=y_his-y_los;
    yc=y_hic-y_loc;
    uint i;
    y.resize(x);
    for(i=0;i<x.N;i++){
      y.elem(i) = sqrt(ys.elem(i)*ys.elem(i)+yc.elem(i)*yc.elem(i));
    }
  }

  extern Parameter <uint> _annealingRepetitions_;
  extern Parameter <double> _annealingCooling_;

  /*! the Hamming distance between two matrices #fix# and #fox# when
      the nodes of #fox# are permuted according to #p#. If they don't
      match in size and #sub=1#, then only the upper left submatrices
      are compared; if #sub=0#, the missing entries of the smaller
      matrix are counted as wrong symbols */
  template<class X>
  double matdistance(Array<X>& fix,Array<X>& fox,Permutation& p,bool sub){
    CHECK(fox.d0==p.N,"matrix and its permutation don't agree in size");
    uint i,j,Nmin,Nmax,n=0;
    if(fix.d0<=fox.d0){ Nmin=fix.d0; Nmax=fox.d0; }
    else{ Nmin=fox.d0; Nmax=fix.d0; }
    for(i=0;i<Nmin;i++) for(j=0;j<Nmin;j++) if(fix(i,j)!=fox(p(i),p(j))) n++;
    if(!sub) n+=(Nmax-Nmin)*(Nmax+Nmin);
    //std::cout <<fix <<fox <<n;
    return ((double)n)/Nmax/Nmax;
  }

  /*! same as #matdistance(A,B,identical-permutation,sub)# */
  template<class X>
  double matdistance(Array<X>& A,Array<X>& B,bool sub){
    Permutation p;
    p.init(B.d0);
    return matdistance(A,B,p,sub);
  }

  /* Uses simulated annealing to find the permutation #p# such that
     #matdistance(fix,fox,p,sub)# becomes minimal */
  template<class X>
  double matannealing(Array<X>& fix,Array<X>& fox,Permutation& p,bool sub){
    CHECKO(fix.nd==2 && fox.nd==2);

    uint N=fox.d0;
    p.init(N); p.store();
    if(fox.N<=1 && fix.N<1) return matdistance(fix,fox,p,sub);
  
    uint i,j,k;
    unsigned long t;
    double newdist,dist,bestdist=1,temp;
  
    for(k=0;k<_annealingRepetitions_;k++){
      t=N*N/10;
      p.random();
      temp=0.1;
      dist=matdistance(fix,fox,p,sub);
      if(dist<bestdist){ if(dist==0) return 0; bestdist=dist; p.store(); }
      while(temp>0.001){
	i=rnd(N); j=rnd(N-1); if(j>=i) j++;
	p.permute(i,j);
	newdist=matdistance(fix,fox,p,sub);
	if(::exp(-(newdist-dist)/temp)>rnd.uni()){
	  dist=newdist;
	  if(dist<bestdist){
	    if(dist==0) return 0;
	    bestdist=dist;
	    p.store();
	  }
	}else{
	  p.permute(i,j);
	}
	if(!(t--)){ t=N*N/10; temp*=_annealingCooling_; }
      }
    }
    p.restore();
    return bestdist;
  }
}

/*! a solver for monotonic unimodal 1D functions */
class MonSolver{
public:
  double min,max;
  int phase;
  MonSolver(){
    min=-.1; max=.1;
    phase=0;
  }
  void init(double& par,double wide=2.){
    phase=0;
    double w=.5*wide*(max-min);
    max+=w;
    min-=w;
    par=min;
  }
  void solve(double& par,const double& err){
    if(phase==0){
      CHECK(par==min,"not phase 0!");
      if(err>0.){
	min-=max-min;
	par=min;
	return;
      }else{
	phase=1;
	par=max;
	return;
      }
    }
    if(phase==1){
      CHECK(par==max,"not phase 1!");
      if(err<0.){
	max+=max-min;
	par=max;
	return;
      }else{
	phase=2;
	par=.5*(max+min);
	return;
      }
    }
    if(phase==2){
      if(err>0.) max=par; else min=par;
      par=.5*(max+min);
    }
  }
};

//===========================================================================
//
// Linear Statistics
//

/*! collects means and (co-)variances of data, also input-output data.
    This class is the basis for linear regression techniques, like
    Least Squares, Partial Least Squares, but can also be used on its
    own, simply to average or calculate (co-)variances \ingroup
    regression */
class LinearStatistics{
public:
  double accum;  //!< the accumulation norm (=number of collected data points if not weighted)
  arr meanX,meanY,varX,covXY; //<! these are not normalized or centered bufferes
  arr MeanX; //!< X mean
  arr MeanY; //!< Y mean
  arr VarX;  //!< X variance
  arr CovXY; //!< XY covariance
  Parameter<double> lambda; //!< forgetting rate [default=0]

  bool computed; //!< internal indicator whether recomputation is needed

  LinearStatistics():
  lambda("LinearStatisticsForgettingRate",0.){ accum=0.; computed=false; }

  /*! NOTE: when data is added to the statistics, the means and
  variances are not yet normalized and centered (computed indicates that) */
  void compute();

  void computeZeroMean();

  /*! add a single new datum (x and y are 1D arrays)
      or a batch of data (x and y are 2D arrays) to the statistics in a weighed fashion */
  void learn(const arr& x,const arr& y,double weight=1.);

  //! in case y is a scalar
  void learn(const arr& x,double y,double weight){ learn(x,arr(&y,1),weight); }

  //! in case there is no y
  void learn(const arr& x){ if(x.nd==1) learn(x,arr(0u),1.); else learn(x,arr(x.d0,0u),1.); }

  //! weighted learning
  void learn(const arr& x,double weight){ if(x.nd==1) learn(x,arr(0u),weight); else learn(x,arr(x.d0,0u),weight); }

  void clear(){ forget(1); }

  //! forget the current statistics partially
  void forget(double lambda=1.);

  uint inDim(){ return meanX.N; }

  double variance(){ compute(); return trace(VarX); }

  void correlationX(arr& corr,bool clearDiag=false){
    compute();
    corr=VarX;
    arr sd(corr.d0);
    uint i,j;
    for(i=0;i<corr.d0;i++) sd(i)=::sqrt(corr(i,i));
    for(i=0;i<corr.d0;i++) for(j=0;j<corr.d1;j++)
      if(sd(i)*sd(j)) corr(i,j)/=sd(i)*sd(j); else corr(i,j)=0.;
    if(clearDiag) for(i=0;i<corr.d0;i++) corr(i,i)=0.;
  }

  void correlationXY(arr& corr,bool clearDiag=false){
    compute();
    corr=CovXY;
    arr sd(corr.d0);
    uint i,j;
    for(i=0;i<corr.d0;i++) sd(i)=::sqrt(corr(i,i));
    for(i=0;i<corr.d0;i++) for(j=0;j<corr.d1;j++)
      if(sd(i)*sd(j)) corr(i,j)/=sd(i)*sd(j); else corr(i,j)=0.;
    if(clearDiag) for(i=0;i<corr.d0;i++) corr(i,i)=0.;
  }

  //! returns the mahalanobis metric relative to the VarX matrix
  void mahalanobisMetric(arr& metric){
    compute();
    metric.resize(VarX);
    arr var;
    var=VarX;
    for(uint i=0;i<var.d0;i++) var(i,i)+=1./(1.+accum);
    ::inverse(metric,var);
  }

  void regressor(arr& A){
    compute();
    A = CovXY * ::inverse(VarX);
  }

  void regressor(arr& A,arr& a){
    compute();
    A = CovXY * ::inverse(VarX);
    a = MeanY - A*MeanX;
  }

  void predict(const arr& x,arr& y){
    compute();
    arr A;
    A = ~CovXY * ::inverse(VarX);
    y = MeanY + A*(x-MeanX);
  }


  //! prototype for operator<<
  void write(std::ostream& os) const{
    os
      <<"<LinearStatistics>"
      <<"\ndim X=" <<MeanX.N
      <<", dim Y=" <<MeanY.N
      <<", accum=" <<accum
      <<"\nmean X=" <<MeanX.ioraw()
      <<"\nmean Y=" <<MeanY
      <<"\nvariance X="<<VarX
      <<"\ncovariance XY=" <<CovXY
      <<"\n</LinearStatistics>"
      <<std::endl;
  }
};
stdOutPipe(LinearStatistics);

/*! Rprop, a fast gradient adaptation */
class Rprop{
public:
  double incr;
  double decr;
  double dMax;
  double dMin;
  double rMax;
  
  arr lastGrad; // last error gradient
  arr stepSize; // last update

  Rprop(){
    incr   = 1.2;
    decr   = .33;
    dMax = 50;
    dMin = 1e-6;
    rMax = 0;
  }

  void init(uint size,double delta0 = 1.) { 
    stepSize.resize (size);
    lastGrad.resize (size);
    stepSize  = delta0; 
    lastGrad  = 0; 
  }
  void forget(){ init(stepSize.N); }
  bool fine(){ double maxStep = stepSize(stepSize.maxIndex()); return maxStep < incr*dMin; }
  void learn(arr& w,const arr& grad){
    if(!stepSize.N) init(grad.N);
    CHECK(grad.N==stepSize.N,"gradient dimensionality changed!");

    for(unsigned  i = 0; i < w.N; i++) {
      if(grad.elem(i) * lastGrad(i) > 0){        //same direction as last time
	if(rMax) dMax=fabs(rMax*w.elem(i));
	stepSize(i) = mymin(dMax, incr * stepSize(i)); //increase step size
	w.elem(i) += stepSize(i) * -sgn(grad.elem(i)); //step in right direction
	lastGrad(i) = grad.elem(i);                    //memorize gradient
      }else if(grad.elem(i) * lastGrad(i) < 0){  //change of direction
	stepSize(i) = mymax(dMin, decr * stepSize(i)); //decrease step size
	w.elem(i) += stepSize(i) * -sgn(grad.elem(i)); //step in right direction
	lastGrad(i) = 0;                               //memorize to continue below next time
      }else{                                     //after change of direcion
	w.elem(i) += stepSize(i) * -sgn(grad.elem(i)); //step in right direction 
	lastGrad(i) = grad.elem(i);                    //memorize gradient
      }
    }
  }
  void learn(double& w,const double& grad){
    static arr W,GRAD;
    W.referTo(&w,1); GRAD.referTo(&grad,1);
    learn(W,GRAD);
  }

private:
  int sgn(double x){ if (x > 0) return 1; if (x < 0) return -1; return 0; }
  double mymin(double x,double y){ return x < y ? x : y; }
  double mymax(double x,double y){ return x > y ? x : y; }
};

class TupleIndex:public uintA{
public:
  uintA tri;

  //! initializes to ``k out of n''
  void init(uint k,uint n){
    CHECK(k<=n,"TupleIndex: can't choose "<<k<<" out of "<<n);
    uint i,j;
    tri.resize(n+1,k+1); tri=0;
    for(i=0;i<=n;i++) for(j=0;j<=i && j<=k;j++){
      if(i==0 || j==0 || j==i){
        tri(i,j)=1;
      }else{
	tri(i,j)=tri(i-1,j-1) + tri(i-1,j);
      }
    }

    uint N=tri(n,k);
    resize(N,k);
    for(j=0;j<k;j++) operator()(0,j)=j;
    for(i=1;i<N;i++){
      for(j=0;j<k;j++) operator()(i,j)=operator()(i-1,j);
      for(j--;operator()(i,j)==n+j-k;j--);
      operator()(i,j)++;
      for(j++;j<k;j++) operator()(i,j)=operator()(i,j-1)+1;
    }
  }

  uint index(uintA i){
    uint a=0,n=tri.d0-1,k=i.N,u;
    if(k) a+=tri(n,k)-tri(n-i(0),k);
    for(u=1;u<i.N;u++){
      CHECK(i(u)>i(u-1) && i(u) < n,"wrong symmetic index!");
      a+=tri(n-i(u-1)-1,k-u)-tri(n-i(u),k-u);
    }
    std::cout <<a <<std::endl;
    return a;
  }

  void checkValid(){
    uint i;
    for(i=0;i<d0;i++) CHECK(i==index(operator[](i)),"wrong index association");
  }
};





/*! a Kalman filter */
class Kalman{
  arr
    A,a, //linear forward transition
    Q, //covariance of forward transition x(t) = A*x(t-1) + \NN(0,Q)
    C, //linear observation matrix
    R; //covariance of observation: y = C*x + \NN(0,R)
  
  void setTransitions(uint d,double varT,double varO){
    a.resize(d); a.setZero();
    A.setDiag(1.,d);
    Q.setDiag(varT,d);
    C.setDiag(1.,d);
    R.setDiag(varO,d);
  }

  //notation follows Kevin Murphey's PhD thesis sec 3.6.1 (also in his DBN tutorial)
  void filter(arr& Y,arr& X,arr& V,arr *Rt=0){
    CHECK(Y.nd==2,"");
    uint d=Y.d1,n=Y.d0,t;
    CHECK(!Rt || (Rt->nd==3 && Rt->d0==n && Rt->d1==d && Rt->d2==d) ,"");
    
    arr e,S,K,L,Id;
    X.resize(n,d);
    V.resize(n,d,d);
    Id.setId(d);
    
    //initialization:
    X[0] = inverse(C) * Y[0];
    if(Rt) R=(*Rt)[0];
    V[0] = ~C * R * C + Q;

    for(t=1;t<n;t++){
      X[t] = A * X[t-1]; //mean: forward predicted
      V[t] = A * V[t-1] * ~A + Q; //variance: forward predicted
      e    = Y[t] - C * X[t]; //observation error w.r.t. prediction
      if(Rt) R=(*Rt)[t];
      S    = C * V[t] * ~C + R; //variance of observation error
      K    = V[t] * ~C * inverse(S); //Kalman gain
      //L  = ::log \NN(e ; 0,S ) //observation log-likelihood
      X[t] = X[t] + K * e;
      V[t] = (Id - K*C) * V[t]; //= V[t] - K * S * ~K;
    }
  }

  //notation follows Kevin Murphey's PhD thesis sec 3.6.1 (also in his DBN tutorial)
  void smooth(arr& Y,arr& X,arr& V,arr *Vxx=0,arr *Rt=0){
    CHECK(Y.nd==2,"");
    uint d=Y.d1,n=Y.d0,t;
    CHECK(!Rt || (Rt->nd==3 && Rt->d0==n && Rt->d1==d && Rt->d2==d) ,"");
    
    arr e,S,K,L,Id;
    X.resize(n,d);
    V.resize(n,d,d);
    Id.setId(d);
    
    //initialization:
    X[0] = inverse(C) * Y[0];
    if(Rt) R=(*Rt)[0];
    V[0] = ~C * R * C + Q;

    for(t=1;t<n;t++){
      X[t] = A * X[t-1] + a; //mean: forward predicted
      V[t] = A * V[t-1] * ~A + Q; //variance: forward predicted
      e    = Y[t] - C * X[t]; //observation error w.r.t. prediction
      if(Rt) R=(*Rt)[t];
      S    = C * V[t] * ~C + R; //variance of observation error
      K    = V[t] * ~C * inverse(S); //Kalman gain
      //L  = ::log \NN(e ; 0,S ) //observation log-likelihood
      X[t] = X[t] + K * e;
      V[t] = (Id - K*C) * V[t]; //= V[t] - K * S * ~K;
    }

    arr xp,Vp,J;
    if(Vxx) Vxx->resize(n-1,d,d);
    for(t=n-1;t--;){
      xp   = A * X[t] + a; //mean: forward predicted
      Vp   = A * V[t] * ~A + Q; //variance: forward predicted
      J    = V[t] * ~A * inverse(Vp); //smoother gain matrix
      X[t] = X[t] + J * (X[t+1] - xp);
      V[t] = V[t] + J * (V[t+1] - Vp) * ~J;
      if(Vxx) (*Vxx)[t] = J * V[t+1]; //is cross covariance between x(t) and x(t+1)
    }
  }

  void EMupdate(arr& Y,arr *Rt=0){
 uint n=Y.d0,t; // d=Y.d1;
    arr X,V,Vxx;

    smooth(Y,X,V,&Vxx,Rt);

#if 0
    arr xx(d,d),vv(d,d),corr(d,d),sd(d); xx.setZero(); vv.setZero();
    for(t=0;t<n-1;t++){
      xx += Vxx[t];// * inverse(V[t]);
      vv += V[t];
    }
    //xx/=n-1.;
    //vv/=n-1.;

    std::cout <<"EM-xx=" <<xx <<vv <<xx * inverse(vv);
#endif

    LinearStatistics S;
    for(t=0;t<n-1;t++) S.learn(X[t],X[t+1]);
    //S.regressor(A);
    S.regressor(A,a);

    //S.forget();
    //for(t=0;t<n-1;t++) S.learn(X[t],Y[t]);
    //S.regressor(C);

    std::cout <<"EM-update: A,a=" <<A <<a <<std::endl; //S.CovXY <<S.VarX <<std::endl;
  }

  void fb(arr& y,arr& f,arr& F,arr& g,arr& G,arr& p,arr& P,arr *Rt=0){
    CHECK(y.nd==2,"");
    uint d=y.d1,n=y.d0,t,T=y.d0;
    CHECK(!Rt || (Rt->nd==3 && Rt->d0==n && Rt->d1==d && Rt->d2==d) ,"");
    
    f.resize(n,d);
    f.resize(n,d,d);

    // Some constant matrices :
    arr &CovH=Q,&CovV=R,&B=C; //rename my matricies
    arr invH,invHA,AinvHA,BinvV,BinvVB,invK,invL,Lconst,J;
    invH=inv(CovH); invHA=invH*A; AinvHA=~A*invHA;
    BinvV=~B*inv(CovV); BinvVB=BinvV*B;
    if(Rt){ BinvV=~B*inv((*Rt)[0]); BinvVB=BinvV*B; }

    // Forward Pass
    F.resize(n,d,d);
    f.resize(n,d);
    F[0] = BinvVB;     //inv(CovP) + BinvVB;
    f[0] = BinvV*y[0]; //inv(CovP)*meanP + BinvV*y[0];
    for(t=1;t<T;t++){
      if(Rt){ BinvV=~B*inv((*Rt)[t]); BinvVB=BinvV*B; }
      invK = inv(AinvHA+F[t-1]);
      F[t] = invH - invHA*invK*~invHA + BinvVB;
      f[t] = invHA*invK*f[t-1] + BinvV*y[t];
    }

    // Backward Pass
    G.resize(n,d,d);
    g.resize(n,d);
    Lconst = invH+BinvVB;
    G[T-1]=0.;
    g[T-1]=0.;
    for(t=T-1;t--;){
      if(Rt){ BinvV=~B*inv((*Rt)[t+1]); BinvVB=BinvV*B; Lconst = invH+BinvVB; }
      invL = inv(Lconst + G[t+1]);
      G[t] = AinvHA - ~invHA*invL*invHA;
      g[t] = ~invHA*invL*(BinvV*y[t+1]+g[t+1]);
    }

    // single marginals:
    P.resize(n,d,d);
    p.resize(n,d);
    for(t=0;t<T;t++){
      P[t] = inv(F[t]+G[t]);
      p[t] = P[t] * (f[t]+g[t]);
    }

    // pair marginals: the means are as above
    arr PP(n-1,2*d,2*d),hh(n,d,d),h1h(n-1,d,d),hh1(n-1,d,d);
    for(t=0;t<T-1;t++){
      if(Rt){ BinvV=~B*inv((*Rt)[t+1]); BinvVB=BinvV*B; }
      J = blockMatrix(F[t]+AinvHA , -~invHA,
		      -invHA      , G[t+1]+invH+BinvVB);
      //jtmp(1:H,1) = f{t}+BinvV*v{t};
      //jtmp(H+1:2*H,1) = g{t};
      PP[t] = inv(J);
      h1h[t]= PP[t].sub(d,2*d-1,0,d-1) + p[t+1]*~p[t]; //lower-left submatrix
      hh1[t]= PP[t].sub(0,d-1,d,2*d-1) + p[t]*~p[t+1]; //upper-right submatrix
    }
    for(t=0;t<T;t++){
      hh[t] = P[t] + p[t]*~p[t];
    }

    // transform fwd & bwd passes to moment representation
    for(t=0;t<T;t++){
      F[t] = inv(F[t]);  f[t] = F[t]*f[t];
      G[t] = inv(G[t]);  g[t] = G[t]*g[t];
    }

    // EM-update:
    arr up(d,d),dn(d,d);
    std::cout <<"EM-update: before:\nA=" <<A <<" B=" <<B <<" covH=" <<CovH <<" covV=" <<CovV <<std::endl;

    for(t=0,up=0.,dn=0.;t<T-1;t++){ up += h1h[t]; dn += hh[t]; }
    A = up * inv(dn);

    for(t=0,up=0.,dn=0.;t<T;t++){ up += y[t]*~p[t]; dn += hh[t]; }
    //B = up * inv(dn);

    for(t=0,up=0.,dn=0.;t<T;t++){ up += y[t]*~y[t] - B*hh[t]*~B; }
    //CovV = up/(double)T;

    for(t=0,up=0.,dn=0.;t<T-1;t++){ up += hh[t+1] - 2.*A*hh1[t] + A*hh[t]*~A; }
    CovH = up/(T-1.);

    std::cout <<"after:\nA=" <<A <<" B=" <<B <<" covH=" <<CovH <<" covV=" <<CovV <<std::endl;

    // log-likelihood:
    double LL=0.,l;
    arr z,Z;
    for(t=0;t<T;t++){
      z = A*p[t];           //mean output
      Z = CovV + A*CovH*~A; //output variance
      z = y[t]-z;
      l=-.5*scalarProduct(Z,z,z) - .5*d*::log(MT_2PI) - .5*::log(determinant(Z));
      LL += l;
    }
    LL/=T;
    std::cout <<"log-likelihood=" <<LL <<std::endl;
  }

private:
  arr inv(const arr& A){ return inverse(A); }
  arr blockMatrix(const arr& A,const arr& B,const arr& C,const arr& D){
    CHECK(A.nd==2 && B.nd==2 && C.nd==2 && D.nd==2,"");
    CHECK(A.d0==B.d0 && A.d1==C.d1 && B.d1==D.d1 && C.d0==D.d0,"");
    uint i,j,a=A.d0,b=A.d1;
    arr X(A.d0+C.d0,A.d1+B.d1);
    for(i=0;i<A.d0;i++) for(j=0;j<A.d1;j++) X(i  ,j  )=A(i,j);
    for(i=0;i<B.d0;i++) for(j=0;j<B.d1;j++) X(i  ,j+b)=B(i,j);
    for(i=0;i<C.d0;i++) for(j=0;j<C.d1;j++) X(i+a,j  )=C(i,j);
    for(i=0;i<D.d0;i++) for(j=0;j<D.d1;j++) X(i+a,j+b)=D(i,j);
    return X;
  }
};


//---------------------------------------------------------------------------
//
// X-Splines
//

class XSpline{
public:
  double DELTA;     // Distance between each
  
  arr V; //vertices
  arr W; //weights associated to each vertex
  
public:
  
  XSpline(){ DELTA = 1.; };
  ~XSpline(){};
  
  void setWeights(double w){ W.resize(V.d0); W=w; W(0)=W(W.N-1)=0.; }
  
  //! suggested: (true,1.) or (false,[.5,1.])
  void type(bool hit,double smooth){ if(hit) setWeights(-smooth); else setWeights(smooth); }
  
  void referTo(arr& P){ V.referTo(P); }
  
  arr eval(double t){ arr x; eval(t,x); return x; }
  void eval(double t,arr& x,arr* v=0);
  void eval(double t,arr& x,arr& v){ eval(t,x,&v); }
  // Hopefully the function below is faster than the 
  // generic pow function for integer powers.
private:
  double intpow(const double x, const int n ){
    double val = x;
    for (int i=1;i<n;i++)
      val *= x;
    return val;
  }
  
  
  // g(u) as defined in the above mentioned article
  double g(double u, double q, double p){
    return 
      q *                            u 
      + 2 * q *                 intpow(u,2)
      + (10 - 12 * q - p) *     intpow(u,3)
      + (2 * p + 14 * q - 15) * intpow(u,4)
      + (6 - 5*q - p) *         intpow(u,5);
  }
  double dgdu(double u, double q, double p){
    return 
      q *                                  1. 
      + 2 * q *                        u    *2.
      + (10 - 12 * q - p) *     intpow(u,2) *3.
      + (2 * p + 14 * q - 15) * intpow(u,3) *4.
      + (6 - 5*q - p) *         intpow(u,4) *5.;
  }
  // h(u)
  double h(double u, double q){
    return 
      q *            u
      + 2 * q * intpow(u,2)  
      - 2 * q * intpow(u,4)
      - q *     intpow(u,5);
  }
  
  double dhdu(double u, double q){
    return 
      q *                  1.
      + 2 * q *        u    *2.  
      - 2 * q * intpow(u,3) *4.
      - q *     intpow(u,4) *5.;
  }
};



//===========================================================================
//
// Partial Least Squares (PLS, SIMPLS, de Jong)
//

/*! An implementation of Partial Least Squares following the SIMPLS algorithm (de Jong).
    This PLS implementation finds an optimal linear regression (from n to m dimensions)
    by first calculating input projections `with higest correlation to the output'. This
    implementation uses the Singular Value Decomposition routine MT::svd and
    the LinearStatistics. \ingroup regression */
class PartialLeastSquares{
public:
  LinearStatistics S; //<! the statistics collected with learning

  arr W,Q,B; //the transformation matricies

  arr resErr; //the residual errors

  Parameter<uint> maxProj;

  PartialLeastSquares():maxProj("PLSmaxProjections",0){ }

  //! add a new datum to the statistics -- calls LinearStatistics::learn (doesn't execute PLS yet!)
  void learn(const arr& x,const arr& y,double weight=1.){ S.learn(x,y,weight); }

  //! same for 1D output
  void learn(const arr& x,double y,double weight=1.){ S.learn(x,y,weight); }

  void clear(){ S.clear(); }

  //! executes PLS, computing the linear mapping from the projections
  void SIMPLS();

  //! the learned mapping
  void map(const arr& x,arr& y){
    if(!S.computed) SIMPLS();
    if(x.nd==2){
      y.resize(x.d0,outDim());
      for(uint i=0;i<x.d0;i++) map(x[i],y[i]());
      return;
    }
    arr xx;
    xx=x-S.MeanX;
    y=B*xx;
    y+=S.MeanY;
    //y=B*(x-MeanX)+MeanY;
  }

  //! in case y is a scalar
  void map(const arr& x,double& y){ arr _y(&y,1); map(x,_y); }

  //! return the output scalar
  double map(const arr& x){ double o; map(x,o); return o; }

  //! return the k-th projection learned by PLS
  arr projection(uint k){
    CHECK(k<W.d0,"PLS: don't have "<<k<<"th projection");
    return W[k];
  }

  uint inDim(){ return S.meanX.N; }
  uint outDim(){ return S.meanY.N; }

  void write(std::ostream& os) const{ os <<S; }
};


//---------------------------------------------------------------------------
//
// Approximate Nearest Neighbor Search (kd-trees, etc)
//

#ifdef MT_ANN
#include "list.h"
#include<ANN/ANN.h>

struct doublePair{ arr x,y; };
std::ostream& operator<<(std::ostream& os,const doublePair& p){ os <<p.x <<p.y; return os; }

class ANN{
public:
  MT::List<doublePair> X;
  MT::Array<double*> C;
  intA L;
  uint D,Dy,treeS;

private:
  ANNbd_tree *tree;
  PartialLeastSquares pls;

public:

  ANN(){ tree=0; treeS=0; D=Dy=0; }
  ~ANN(){ if(tree) delete tree; annClose(); }

  void learn(const arr& x,const arr& y){ add(x,&y); }

  void add(const arr& x,const arr* y=0){
    if(x.nd==2){ for(uint i=0;i<x.d0;i++) add(x[i]); return; }
    if(!D) D=x.N;
    if(y && !Dy) Dy=y->N;
    CHECK(D==x.N,"");
    CHECK((!y && Dy==0) || (y && Dy==y->N),"");

    //find closest so far
    int l;
    double d;
    if(tree){
      tree->annkSearch(x.p,1,&l,&d,0.);
      if(d<1e-20){ MT_MSG("already added! - yet not handled case!"); return; }
    }

    //add to data
    doublePair p; p.x.referTo(x); if(y) p.y.referTo(*y);
    X.append(p);
    L.append(-1);
    if(tree){
      if(true || X.N-treeS > .1*X.N){
	L=-1;
	delete tree;
	tree=0;
      }else{
	while(L(l)!=-1) l=L(l);
	L(l)=X.N-1;
      }
    }
  }

  void calculate(){
    if(tree) return;
    C.resize(X.N);
    for(uint i=0;i<X.N;i++) C(i)=X(i)->x.p;
    tree = new ANNbd_tree(C.p,X.N,D);
    treeS=X.N;
  }

  void getNN(const arr& x,uint k,intA& idx,double eps=.0,bool talk=false){
    CHECK(X.N>=k,"data has less ("<<X.N<<") than k="<<k<<" points");
    CHECK(x.N==D,"query point has wrong dimension");
    calculate();
    arr dists(k);
    idx.resize(k);
    tree->annkSearch(x.p,k,idx.p,dists.p,eps);
    uint i;
    int l;
    for(i=0;i<k;i++){
      l=L(idx(i));
      while(l!=-1){
	idx.append(l);
	dists.append(sqrDistance(x,X(l)->x));
	l=L(l);
      }
    }
    if(talk){
      std::cout
	<<"ANN query:\n query point "<<x
	<<"\n found neighbors:\n";
      for(uint i=0;i<idx.N;i++){
	std::cout <<' '
	  <<i <<' '
	  <<idx(i) <<'\t'
	  <<sqrt(dists(i)) <<'\t'
	  <<X(idx(i))->x <<std::endl;
      }
    }
  }

  void getNN(arr& x,uint k,arr& xx,double eps=.0,bool talk=false){
    intA idx;
    getNN(x,k,idx,eps,talk);
    xx.resize(idx.N,D);
    for(uint i=0;i<idx.N;i++) xx[i]=X(idx(i))->x;
  }

  void map(const arr& x,arr& y){
    if(x.nd==2){ y.resize(x.d0,Dy); for(uint k=0;k<x.d0;k++) map(x[k],y[k]()); return; }
    if(!X.N) return;
    calculate();
    uint nn=20;
    if(nn>treeS) nn=treeS;
    arr dists(nn);
    intA idx(nn);
    tree->annkSearch(x.p,nn,idx.p,dists.p,0.);
    LinearStatistics S;
    uint i;
    for(i=0;i<nn;i++) S.learn(X(idx(i))->x-x);
    //S.computeZeroMean();
    double d=.5*S.variance(),w;
    if(!d) d=1.;
    pls.S.forget();
    S.forget();
    for(i=0;i<nn;i++){
      w=::exp(-.5*dists(i)/d);
      w=1./(nn+i);
      S.learn(X(idx(i))->y,w);
      //pls.learn(X(idx(i))->x,X(idx(i))->y,w);
    }
    S.compute();
    y=S.MeanX;
    //pls.map(x,y);
  }
};
#endif


#ifdef MT_IMPLEMENTATION
#  include"algos.cpp"
#  include"conjGrad.c"
namespace MT{
  Parameter <uint> _annealingRepetitions_ ("annealingRepetitions",100);
  Parameter <double> _annealingCooling_ ("annealingCooling",.9);
}
#endif

#endif
