/*  Copyright (C) 2000, 2006  Marc Toussaint (mtoussai@inf.ed.ac.uk)
    under the terms of the GNU LGPL (http://www.gnu.org/copyleft/lesser.html)
    see the `std.h' file for a full copyright statement  */

#ifndef MT_dynamics_h
#define MT_dynamics_h

#include "std.h"
#include "array.h"
#include "algos.h"
//#include "linearModels.h"
#include "list.h"

namespace MT{
  double phaseDist(double x0,double v0,double x1,double v1);
  
  double phaseAccel(double x0,double v0,double x1,double v1,double maxA=0);
  
  void phaseTangent(double x0,double v0,double x1,double v1,double& dx,double& dv,double maxA=0);

  inline void oscTrajectory(doubleA& plan,uint T,uint n){
    plan.resize(T,n);
    doubleA x,tau,psi;
    uint t,i;
    x.resize(n);
    tau.resize(n);
    for(i=0;i<n;i++) tau(i)=10.*rnd.gauss()+200;
    for(t=0;t<T;t++) for(i=0;i<n;i++){
      plan(t,i)=::cos(MT_2PI*t/tau(i));
    }
  }

  void desAccel(doubleA& x,doubleA& v,doubleA& x0,doubleA& v0,doubleA& a0,doubleA& desAcc,double deltaT=.01,double decaySteps=10.,double oscillations=0.){
    CHECK(x.N==v.N &&
      x0.N==v0.N &&
      x.N==x0.N,"state or plan have wrong dimensionality");
    
    //gains
    double a,b,c;
    b=2./(decaySteps*deltaT);
    //a=(1.+oscillations)/(decaySteps*deltaT); a=a*a;
    a=b*b*(MT_PI*MT_PI*oscillations*oscillations + .25);
    c=.75;//c=.0;

    double maxAcc=.2/deltaT;
    uint i;
    
    desAcc.resize(x);
    for(i=0;i<x.N;i++){
      desAcc(i) = a*(x0(i)-x(i));
      desAcc(i)+= b*(v0(i)-v(i));
      desAcc(i)+= c*a0(i);
      
      if(fabs(desAcc(i))>maxAcc) desAcc(i)=maxAcc*sign(desAcc(i));
    }
  }

  struct Controller{
    double tau,a,b,c,maxAcc,initMass;
    doubleA lastV,lastV0,lastF,lastDA,desAcc,mass;
    Rprop rprop;
    bool adapt,smooth;

    void init(double deltaT,double decaySteps,double oscillations,double mass){
      tau=deltaT;
      b=2./(decaySteps*tau);
      //a=(1.+oscillations)/(decaySteps*tau); a=a*a;
      a=b*b*(MT_PI*MT_PI*oscillations*oscillations + .25);
      c=.75; //c=.0;
      maxAcc=.1/tau;
      initMass=mass;
      if(!initMass) initMass=1e-10;
      adapt=true;
      smooth=false;
    }
    
    void pid(doubleA& x,doubleA& v,doubleA& plan,uint t,doubleA& f){
      doubleA x0(plan,t),x1(plan,t+1),x2(plan,t+2),v0,v1;
      v0=x1-x0; v0/=tau;
      pid(x,v,x0,v0,f);
    }

    void pid(doubleA& x,doubleA& v,doubleA& x0,doubleA& v0,doubleA& f){
      CHECK(x.N==v.N &&
	x0.N==v0.N &&
	x.N==x0.N,"state or plan have wrong dimensionality");

      uint i;

      if(!mass.N){ mass.resize(2,x.N); mass[0]=initMass; mass[1]=0.; }

      //adapt/learn linear relation between accelerations and forces
      if(adapt){
	if(!rprop.stepSize.N){ //first time!
	  rprop.init(mass.N,.1*initMass);
	  rprop.incr=1.1;
	  rprop.decr=.8;
	  rprop.rMax=.01;
	}else{
          doubleA trueAcc;
	  if(lastV.N){ trueAcc=v-lastV; trueAcc/=tau; }else{ trueAcc.resize(x); trueAcc.setZero(); }
	  doubleA grad; grad.resize(mass);
	  for(i=0;i<x.N;i++){
	    grad(0,i) = - lastF(i) * (desAcc(i) - trueAcc(i));
	    grad(1,i) = -            (desAcc(i) - trueAcc(i));
	  }
	  rprop.learn(mass,grad);
	  for(i=0;i<x.N;i++) if(mass(0,i)<0.){ mass(0,i) = 1e-10; }
	  //std::cout <<"mass=" <<mass[0] <<" grav=" <<mass[1] <<std::endl;
	}
      }

      //compute desired accelerations
      desAccel(x,v,x0,v0,desAcc);

      //linearly map desired accelerations to forces
      f.resize(x);
      for(i=0;i<x.N;i++){
        if(smooth && lastDA.N){
	  desAcc(i)=.7*lastDA(i) + .3*desAcc(i);
	  /*
	  if(desAcc(i)*lastDA(i)>0.){
	    if(desAcc(i)/lastDA(i)>1.2) desAcc(i)=1.2*lastDA(i);
	    //if(desAcc(i)/lastDA(i)<.5) desAcc(i)=.5*lastDA(i);
	  }else{
	    if(-desAcc(i)/lastDA(i)>1.2) desAcc(i)=-1.2*lastDA(i);
	  }
	  */
	}
	f(i)=mass(0,i)*(desAcc(i) + mass(1,i));
      }

      lastF=f;
      lastV=v;
      lastV0=v0;
      lastDA=desAcc;
    }

    void desAccel(doubleA& x,doubleA& v,doubleA& x0,doubleA& v0,doubleA& desAcc){
      CHECK(x.N==v.N &&
	x0.N==v0.N &&
	x.N==x0.N,"state or plan have wrong dimensionality");

      uint i;
      doubleA a0;

      // desired idealistic (null-error) acceleration
      if(lastV0.N){ a0=v0-lastV0; a0/=tau; }else{ a0.resize(x); a0.setZero(); }

      desAcc.resize(x);
      for(i=0;i<x.N;i++){
	desAcc(i) = a*(x0(i)-x(i));
	desAcc(i)+= b*(v0(i)-v(i));
	desAcc(i)+= c*a0(i);

	if(fabs(desAcc(i))>maxAcc) desAcc(i)=maxAcc*sign(desAcc(i));
      }

      lastV=v;
      lastV0=v0;
    }

    void write(std::ostream& os) const{ os <<mass; }
    void read(std::istream& is){ is >>mass; }
  };
  
  struct FreeMass{
    doubleA x,v,m,g;
    void init(uint n){
      x.resize(n); v.resize(n); m.resize(n); g.resize(n);
      x=v=0;
      rndUni(m,.1,10);
      rndUni(g,-10.,10.);
      //g=0.;
      std::cout <<m <<' ' <<g <<std::endl;
      x=0.;
    }
    void step(doubleA& f,double tau=.01){
      if(!x.N) init(f.N);
      for(uint i=0;i<v.N;i++){
	v(i)+=tau*(g(i)+f(i)/m(i));
        x(i)+=tau*v(i);
      }
    }

  };

  class Attractor{
  public:
    struct Peak{ doubleA x; double w,sig; };
    MT::List<Peak> peaks;
    typedef MT::List<Peak>::node node;
    uint D;
    doubleA pos,field,gradient;
    
    Attractor(){ D=0; }
    
    void setD(uint d){
      CHECK(!D || D==d,"can't change dimensionality");
      D=d;
      pos.resize(d); pos.setZero();
    }
    
    void addPeak(double x,double y,double w,double sig){
      setD(2);
      node n=peaks.new_node();
      n->x.resize(2); n->x(0)=x; n->x(1)=y;
      n->w=w;
      n->sig=sig;
    }

    void addPeak(doubleA& x,double w,double sig){
      setD(x.N);
      node n=peaks.new_node();
      n->x=x;
      n->w=w;
      n->sig=sig;
    }

    double getValue(doubleA& x){
      double v=0.,d;
      node n;
      forall(n,peaks){
	d=norm(n->x-x)/n->sig;
        v += n->w*::exp(-.5*d*d);
      }
      return v;
    }
    
    void getGradient(){
      gradient.resize(D); gradient.setZero();
      node n;
      double d;
      forall(n,peaks){
	d=norm(n->x-pos)/n->sig;
	gradient += (n->w*::exp(-.5*d*d)) * (n->x-pos);
      }
    }
    
    node getMax(){
      node n,max=0;
      forall(n,peaks) if(!max || n->w>max->w) max=n;
      return max;
    }

    void getGradientToBestValue(){
      double v,maxV;
      node n,max=0;
      forall(n,peaks){
	v=n->w/(10.+norm(n->x-pos)/n->sig);
	if(!max || v>maxV){ max=n; maxV=v; }
      }
      if(!max) gradient.setZero();
      gradient=max->x-pos;
    }

    
    void step(double tau=.1,double noise=.001,double max=.0){
      rndGauss(pos,noise,true);
      //getGradient();
      getGradientToBestValue();
      if(max){
        double l=norm(gradient); if(l>max) gradient *= max/l;
      }
      pos += tau * gradient;
    }

    void random(uint d,uint n){
      doubleA x(d);
      uint i;
      for(i=0;i<n;i++){
	rndUni(x,-1,1);
	addPeak(x,rnd.uni(.0,.5),rnd.uni(.1,.5));
      }
    }
    void randomMovePeaks(){
      node n;
      forall(n,peaks){
	rndGauss(n->x,.01,true);
	n->w+=.01*rnd.gauss();
      }
    }
    
    void calcField(uint res,double mi=0.,double ma=0.){
      CHECK(D,"no dimensionality");
      uint i,j;
      double x,y,dx,dy;
      doubleA mini(D),maxi(D);
      mini=maxi=pos;
      node n;
      if(mi && ma){
	for(i=0;i<D;i++){ mini(i)=mi; maxi(i)=ma; }
      }else{
	forall(n,peaks) for(i=0;i<D;i++){
	  if(n->x(i)-3*n->sig<mini(i)) mini(i)=n->x(i)-3*n->sig; if(n->x(i)+3*n->sig>maxi(i)) maxi(i)=n->x(i)+3*n->sig;
	}
      }
      
      if(D==1){
	field.resize(res);
	for(i=0;i<res;i++){
	  x=mini(0) + (maxi(0)-mini(0))*i/(res-1);
	  field(i)=0.;
	  forall(n,peaks){
	    dx=x-n->x(0); dx/=n->sig;
	    field(i)+=n->w*::exp(-.5*dx*dx);
	  }
	}
	return;
      }
      if(D==2){
	field.resize(res,res);
	for(i=0;i<res;i++) for(j=0;j<res;j++){
	  x=mini(0) + (maxi(0)-mini(0))*i/(res-1);
	  y=mini(1) + (maxi(1)-mini(1))*j/(res-1);
	  field(i,j)=0.;
	  forall(n,peaks){
	    dx=x-n->x(0); dx/=n->sig;
	    dy=y-n->x(1); dy/=n->sig;
	    field(i,j)+=n->w*::exp(-.5*(dx*dx+dy*dy));
	  }
	}
	return;
      }
      HALT("can't display that dimension");
    }
  };
}

stdPipes(MT::Controller);

#ifdef MT_IMPLEMENTATION
#  include"dynamics.cpp"
#endif

#endif
