/*  Copyright (C) 2000, 2006  Marc Toussaint (mtoussai@inf.ed.ac.uk)
    under the terms of the GNU LGPL (http://www.gnu.org/copyleft/lesser.html)
    see the `std.h' file for a full copyright statement  */

#ifndef MT_gauss_h
#define MT_gauss_h

#include "array.h"
#include "functions.h"

typedef void (*Trans)(arr& x);

/*! Gaussian \ingroup infer1 */
struct Gaussian{
  arr c; //!< center (normal representation)
  arr C; //!< covariance (normal representation)
  arr u; //!< `center-vector' in canonical representation
  arr U; //!< precision matrix (canonical representation)

  bool okC; //!< flag if normal representation is available
  bool okU; //!< flag if canonical representation is available

  //! okC=okU=false
  Gaussian(){ setCU(false,false); }

  //! dimensionality
  uint N() const{ CHECK(okC || okU,"non-initialized Gaussian (no C or U)"); if(okC) return c.N; else return u.N; }
  //! make canonical representation available
  void makeU() const; //make sure the canonical rep is ok
  //! make normal representation available
  void makeC() const; //make sure the normal rep is ok
  void setCU(bool okc,bool oku){ okC=okc; okU=oku; }

  //! copy operator
  void operator=(const Gaussian& x);

  void setC(const arr& cc,const arr& CC){ c=cc; C=CC; setCU(true,false); }
  void setU(const arr& uu,const arr& UU){ u=uu; U=UU; setCU(false,true); NIY; }
  void setUniform(uint n){
#if 0
    u.resize(n); u.setZero();
    U.resize(n,n); U.setZero();
    setCU(false,true);
#else
    setDiagonal(n,1e10);
#endif
  }
  void setDelta(const arr& e){
    c=e;
    C.setDiag(1e-9,e.N);
    //C.setDiag(0.,e.N);
    setCU(true,false);
    makeU();
  }
  void setRandom(uint n){
    c.resize(n); rndUni(c);
    C.resize(n,n); rndUni(C,1e-3,1.); C=C*~C;
    setCU(true,false);
  }

  void setConditional(uint n,double var2){
    u.resize(2*n); u.setZero();
    U.resize(2*n,2*n); U.setZero();
    for(uint i=0;i<n;i++) U(i+n,i+n) = 1./var2;
    setCU(false,true);
    NIY;
  }
  void setConditional(uint n,double var1,double var2);
  void setConditional(const arr& f,const arr& F,const arr& Q);
  void setDiagonal(uint n,double var);
  void setMarginal(uint n,const Gaussian& m,uintA list){
#if 1
    m.makeU();
    u.resize(n); u.setZero();
    U.resize(n,n); U.setZero();
    uint i,j;
    for(i=0;i<list.N;i++) u(list(i))=m.u(i);
    for(i=0;i<list.N;i++) for(j=0;j<list.N;j++) U(list(i),list(j))=m.U(i,j);
    setCU(false,true);
#else
    m.makeC();
    setDiagonal(n,1e10);
    uint i,j;
    for(i=0;i<list.N;i++) c(list(i))=m.c(i);
    for(i=0;i<list.N;i++) for(j=0;j<list.N;j++) C(list(i),list(j))=m.C(i,j);
    setCU(true,false);
#endif
  }

  void write(std::ostream& os) const;
  void read(std::istream& os);
};
stdOutPipe(Gaussian);
stdInPipe(Gaussian);
stdEqPipe(Gaussian);

typedef MT::Array<Gaussian> gaussA;

//! estimate a gaussian from a table of data points \ingroup infer1
void estimate(Gaussian &g,const arr& X);

/*! collapses a mixture of gaussians (or weighted gaussians: P doesn't
    have to be normalized) to a single gaussian via moment matching \ingroup infer1 */
void collapseMoG(Gaussian& g,const arr& P,const gaussA& G);

/*! similar to the unscented transform: first take systematically samples
    from the Gaussian, then weight the samples proportional to f(x),
    then reestimate the Gaussian \ingroup infer1 */
void resampleAndEstimate(Gaussian& g,double (*f)(const arr& x));

//! generate a single sample from a gaussian \ingroup infer1
void sample(arr& x,const Gaussian &g);

//! generate a table of N samples from a gaussian \ingroup infer1
void sample(arr& X,uint N,const Gaussian &g);

void systematicWeightedSamples(arr& X,arr& W,const Gaussian& g);

//! check if a and b are the same gaussian [approximately - prelim] \ingroup infer1
bool sameGaussian(const Gaussian &a,const Gaussian &b);

//! symmetric Kullback-Leibler divergence \ingroup infer1 
double KLDsym(const Gaussian &a,const Gaussian &b);

/*! map an initial gaussian (a) to a new gaussian (b) via a non-linear function f
    using the unscented transform \ingroup infer1 */
void unscentedTransform(Gaussian &b, const Gaussian &a,Trans f);

//! given x~{a,A} and y|x~{Fx+f,Q}, this returns y~{c,C} \ingroup infer1
void forward(Gaussian& y, const Gaussian& x, arr& f, arr& F, arr& Q);

//! given y~{b,B} and y|x~{Fx+f,Q}, this returns x~{c,C} \ingroup infer1
void backward(Gaussian& x, const Gaussian& y, arr& f, arr& F, arr& Q,double updateStep=1.);

//! given y~{b,B} and y|x~{Fx+f,Q}, this returns x~{c,C}
//void linBwd(Gaussian& x, Gaussian& y, arr& f, arr& F);

//! multiplication: {a,A} * {b,B} = {x,X} * norm \ingroup infer1
void product(Gaussian& x, const Gaussian& a, const Gaussian& b,double *logNorm=0);

//! division: {a,A} / {b,B} = {x,X} + norm \ingroup infer1
void division(Gaussian& x, const Gaussian& a, const Gaussian& b,double *logNorm=0);

//! given x~{a,A} and y|x~{f(x),Q}, this returns y~{c,C} \ingroup infer1
void forward(Gaussian& y, const Gaussian& x, Trans f, arr& Q);

//! given x~{a,A} and y|x~{f(x),Q}, this returns y~{c,C} \ingroup infer1
//void fctFwd(Gaussian& y,Gaussian& x, Trans f);

void getLinFwdFromJoint(arr& f, arr& F, arr& Q, uint n1, uint n2, Gaussian& x);

//! given x~{a,A} and y|x~{Fx+f,Q}, this returns the joint (x,y)~{c,C} \ingroup infer1
void joinMarginalAndConditional(Gaussian& xi,Gaussian& a, arr& f, arr& F, arr& Q);

//! xi is P(x,y), returns P(y|x)={Fx+f,Q}, cut is the lower index of y \ingroup infer1
void getConditional(const Gaussian& xi,uint cut,arr& f,arr& F,arr& Q);

/*! get the marginal of (z_lo,..,z_hi) from a joint over (z_0,..,z_(n-1)) \ingroup infer1 */
void getMarginal(Gaussian& x,const Gaussian& joint,uint lo,uint hi);

/*! get the marginal of (z_list(0),..,z_list(m-1)) from a joint over (z_0,..,z_(n-1)) \ingroup infer1 */
void getMarginal(Gaussian& x,const Gaussian& joint,uintA& list);

/*! -- note, this is not a proper marginal multiplication -- simply set the marginal entries (z_list(0),..,z_list(m-1)) in a joint over (z_0,..,z_(n-1)) \ingroup infer1 */
void setMarginal(Gaussian& joint,const Gaussian& marg,uintA& list);

// assumes middle split
void getMarginalsFromJoint(Gaussian& x,Gaussian& y,Gaussian& xi);

/*! given a joint (x,y)~{c,C} and a potential {b,B} for y|evidence,
    this returns the updated joint (x,y)|evidence~{d,D} \ingroup infer1 */
void multiplyToJoint(Gaussian& xi, Gaussian& b);




void write(const gaussA& x,const char* name);



#ifdef MT_IMPLEMENTATION
#  include "gauss.cpp"
#endif

#endif
