#include<MT/array.h>
#include<MT/algos.h>

#define MT_GP_DEBUG 0
#if MT_GP_DEBUG
#  include<MT/plot.h>
#endif

//===========================================================================
//
//!@name Gaussian Process code
//

struct GaussianProcess{
  //! pointers to data
  doubleA X,Y;
  //! inverse gram matrix (and a second buffer for push/pop)
  doubleA invGram,invGY,ig2;
  //! pointer to kernel function
  double (*kernelF)(void *P,const doubleA& x,const doubleA& y);
  void (*dkernelF)(doubleA& grad,void *P,const doubleA& x,const doubleA& y);
  //! pointer to parameters (a struct or so) passed to the kernel function
  void *kernelP;

  void clear(){ X.resize(0,X.d1); Y.resize(0); invGram.resize(0,0); invGY.resize(0); ig2.resize(0,0); }
  /*! set an arbitrary covariance function,
      P is a pointer to parameters (a struct) that is passed
      everytime when the _kernelF is called */
  void setKernel(void *_kernelP,double (*_kernelF)(void *P,const doubleA& x,const doubleA& y)){
    kernelP=_kernelP;
    kernelF=_kernelF;
    dkernelF=0;
  }
  void setKernel(void *_kernelP,double (*_kernelF)(void *P,const doubleA& x,const doubleA& y),void (*_dkernelF)(doubleA& grad,void *P,const doubleA& x,const doubleA& y)){
    kernelP=_kernelP;
    kernelF=_kernelF;
    dkernelF=_dkernelF;
  }
  //! add a new datum to the data and updates the inv Gram matrix
  void update(const doubleA& x,double& y);
  //! recalculates the inv Gram matrix for the current data
  void calculate();
  //! evaluate the GP at some point - returns y and sig (=standard deviation)
  void evaluate(const doubleA& x,double& y,double& sig);
  //! evaluate the GP at some array of points - returns all y's and sig's
  void evaluate(const doubleA& X,doubleA& Y,doubleA& S);

  void gradient(doubleA& grad,const doubleA& x);

  void push(const doubleA& x,double& y){ ig2=invGram; update(x,y); }
  void pop(){ invGram=ig2; X.resizeCopy(X.d0-1,X.d1); Y.resizeCopy(Y.N-1); }
};


//===========================================================================
//
//!@name standard Gaussian covariance function
//

struct GaussKernelP{
  double noiseVar,priorVar,width;
  GaussKernelP(){ noiseVar=.01; priorVar=.01; width=.04; }
  GaussKernelP(double _noiseSDV,double _priorSDV,double _width){ noiseVar=_noiseSDV*_noiseSDV; priorVar=_priorSDV*_priorSDV; width=_width*_width; }
};

//! you can also pass a double[3] as parameters
double GaussKernel(void *P,const doubleA& x,const doubleA& y){
  GaussKernelP& K = *((GaussKernelP*)P);
  if(&x==&y) return K.priorVar+K.noiseVar;
  double d;
  if(x.N!=1) d=sqrDistance(x,y); else{ d=y(0)-x(0); d=d*d; }
  return K.priorVar*::exp(-.5 * d/K.width);
}

//! you can also pass a double[3] as parameters
void dGaussKernel(doubleA& grad,void *P,const doubleA& x,const doubleA& y){
  GaussKernelP& K = *((GaussKernelP*)P);
  if(&x==&y){ grad.resize(x); grad.setZero(); return; }
  double f=GaussKernel(P,x,y);
  grad = f / K.width * (y-x);
}

double maximizeGP(GaussianProcess& gp,doubleA& x){
  CHECK(gp.Y.N,"");
  if(!x.N){ x.resize(gp.X.d1); x=0.; } //rndGauss(x,1.); }
  doubleA dx;
  Rprop rp;
  rp.init(x.N,.1);
  uint t=0;
  do{
    gp.gradient(dx,x);
    dx *= -1.;
    rp.learn(x,dx);
    t++;
    if(t>100){ cout <<" couldn't stop!"; break; }
  }while(!rp.fine());
  double y,s;
  gp.evaluate(x,y,s);
  cout <<"GP maximization: " <<t <<"iterations, x=" <<x <<" y=" <<y <<" s=" <<s <<endl;
  return y;
}


void randomFunction(GaussianProcess& gp,doubleA& Xbase){
  GaussKernelP& K = *((GaussKernelP*)gp.kernelP);
  double orgNoiseVar=K.noiseVar;
  K.noiseVar=1e-6;

  doubleA x;
  double y,sig;
  uint i;
  //gp.setKernel(&stdKernel,GaussKernel);
  gp.clear();

  Xbase.randomPermute();

#if MT_GP_DEBUG
  doubleA X,Y,S;
  grid(X,1,-5.,5.,1000);
#endif

  //gp.X.resize(0,1); gp.Y.resize(0); //clear current data
  for(i=0;i<Xbase.d0;i++){
#if MT_GP_DEBUG
    gp.evaluate(X,Y,S);
    plotClear();
    plotFunction(X,Y);
    plotFunction(X,Y+S);
    plotFunction(X,Y-S);
    plotPoints(gp.X,gp.Y);
    plot();
#endif
      
    x=Xbase[i]; //get next input point
    gp.evaluate(x,y,sig); //sample it from the GP itself
    y+=sig*rnd.gauss(); //with standard deviation..
    gp.update(x,y);
  }

  K.noiseVar=orgNoiseVar;
}



#ifdef MT_IMPLEMENTATION
#  include"gaussianProcess.cpp"
#endif
